//==========================================================================
//
//      net/tftp_server.c
//
//      Stand-alone TFTP support for RedBoot
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
// Copyright (C) 2002, 2003 Gary Thomas
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    lo
// Contributors: lo
// Date:         2006-02-27
// Purpose:      
// Description:  
//              
// This code is part of RedBoot (tm).
//
//####DESCRIPTIONEND####
//
//==========================================================================

// TFTP server support

#include <redboot.h>     // have_net
#include <net/net.h>
#include <net/tftp.h>
#include <net/tftp_support.h>

externC bool flash_code_overlaps(void *start, void *end);
externC char *flash_errmsg(int err);
externC int flash_erase(void *base, int len, void **err_address);

#define FLASH_SIZE          0x00400000      /* 4M */
#define FLASH_START_ADDR    0xbfc00000
#define KERNEL_FLASH_ADDR   0xbfc80000
#define KERNEL_FLASH_SIZE   0x00110000      /* the size of "kernel" FLASH image */
#define ROOTFS_FLASH_ADDR   0xbfd90000
#define ROOTFS_FLASH_SIZE   0x00250000      /* the size of "rootfs" FLASH image */
#define BOARD_ID_FLASH_ADDR 0xbfc50000

/*#define IMAGE_FILE_LEN_MAX  (KERNEL_FLASH_SIZE+ROOTFS_FLASH_SIZE+256)*/ /* Should be less than 4M */
#define IMAGE_FILE_LEN_MAX  FLASH_SIZE

static struct {
    bool open;
    int  packets_received;
    unsigned long last_good_block;
    int  avail, actual_len;
    struct sockaddr_in serv_addr, cli_addr;
} tftpd_stream;

//char gImgBuf[IMAGE_FILE_LEN_MAX];
char *gImgBuf;
int gImgLen;
//char gTempImgBuf[ROOTFS_FLASH_SIZE]; /* ROOTFS_FLASH_SIZE must be larger than KERNEL_FLASH_SIZE. */
char *gTempImgBuf;
int write_to_flash_ongoing = 0;

static char * errmsg[] = {
  "Undefined error code",               // 0 nothing defined
  "File not found",                     // 1 TFTP_ENOTFOUND 
  "Access violation",                   // 2 TFTP_EACCESS   
  "Disk full or allocation exceeded",   // 3 TFTP_ENOSPACE  
  "Illegal TFTP operation",             // 4 TFTP_EBADOP    
  "Unknown transfer ID",                // 5 TFTP_EBADID    
  "File already exists",                // 6 TFTP_EEXISTS   
  "No such user",                       // 7 TFTP_ENOUSER   
};

int upgrade_firmware_main(void);

/* Send an error packet to the client */
static void 
tftpd_send_error(struct tftphdr * reply, int err)
{
    reply->th_opcode = htons(ERROR);
    reply->th_code = htons(err);
    if ( (0 > err) || ((int)(sizeof(errmsg)/sizeof(errmsg[0])) <= err) )
        err = 0; // Do not copy a random string from hyperspace
    strcpy(reply->th_msg, errmsg[err]);
    	   
    __udp_sendto((char *)reply, 4+strlen(reply->th_msg)+1, &tftpd_stream.cli_addr, &tftpd_stream.serv_addr);
}

//
// Receive a file from the client
//
static int 
tftpd_write_file(struct tftphdr *hdr)
{
    char data_out[SEGSIZE+sizeof(struct tftphdr)];
    char data_in[SEGSIZE+sizeof(struct tftphdr)];
    struct tftphdr *reply = (struct tftphdr *)data_out;
    struct tftphdr *response = (struct tftphdr *)data_in;
    int ok, tries, closed, data_len;
    unsigned short block;
    struct timeval timeout;
    int total_timeouts = 0;

    /*diag_printf("TFTPD(WRQ): %s %s\n", hdr->th_stuff, &hdr->th_stuff[strlen(hdr->th_stuff)+1]);*/
    /*
    if (strncmp(hdr->th_stuff, ".chk", 4)) != 0) {
        tftpd_send_error(reply,TFTP_ENOTFOUND);
        return;
    }
    */

    ok = true;
    closed = false;
    block = 0; /* block number */
    gImgLen = 0;
    gImgBuf = (char *)(((CYG_ADDRWORD)mem_segments[0].start + 0x03ff) & ~0x03ff);
    gTempImgBuf = gImgBuf + FLASH_SIZE;
    /*diag_printf("gImgBuf=%p gTempImgBuf=%p\n", gImgBuf, gTempImgBuf);*/
    while (ok) {
        // Send ACK telling client he can send data
        reply->th_opcode = htons(ACK);
        reply->th_block = htons(block++); // postincrement
        for (tries = 0;  tries < TFTP_RETRIES_MAX;  tries++) {
            if (__udp_sendto((char *)reply, 4, &tftpd_stream.cli_addr, &tftpd_stream.serv_addr) < 0) {
                // Problem sending ACK
                diag_printf("TFTPD: error to send ACK\n");
                //*err = TFTP_NETERR;
                continue; // retry the send ACK
            }
    
        repeat_select:
            timeout.tv_sec = TFTP_TIMEOUT_PERIOD;
            timeout.tv_usec = 0;
            data_len = sizeof(data_in);
            data_len = __udp_recvfrom(data_in, data_len, &tftpd_stream.cli_addr,
                                       &tftpd_stream.serv_addr, &timeout);

            if ( data_len < 0) {
                diag_printf("TFTPD timed out 2\n");
                if (++total_timeouts > TFTP_TIMEOUT_MAX) {
                    tftpd_send_error(reply,TFTP_EBADOP);
                    ok = false;
                    break; // out of the retries loop
                }
                continue; // retry the send, using up one retry.
            }
            else {
                /*
                diag_printf("TFTPD: received %x from %s:%d\n", 
                        ntohs(response->th_opcode), inet_ntoa(&tftpd_stream.cli_addr.sin_addr), tftpd_stream.cli_addr.sin_port);
                if (ntohs(response->th_opcode) == DATA)
                {
                    diag_printf("Block#: %d DataLen:%d\n",ntohs(response->th_block), data_len-4);
                }
                */
                if ((block%128) == 0)
                    sysGpioSet(7, 0); /* turn off Test LED */
                else if ((block%64) == 0)
                    sysGpioSet(7, 1); /* turn on Test LED */
                /* 
                 * Print one dot character whenever receiving 1K bytes,
                 * and 64 dot characters (64K) per line.
                 */
                if ((block%2) == 0)
                    diag_printf(".");
                if ((block%128) == 0)
                    diag_printf("\n");
                    
                if ((ntohs(response->th_opcode) == DATA) &&
                    (ntohs(response->th_block) < block)) {
    
                    // Then it is repeat DATA with an old block; listen again,
                    // but do not repeat sending the current ack, and do not
                    // use up a retry count.  (we do re-send the ack if
                    // subsequently we time out)
                    goto repeat_select;
                }
            
                if ((ntohs(response->th_opcode) == DATA) &&
                    (ntohs(response->th_block) == block)) {
                    // Good data - write to file
                    if((gImgLen+data_len-4)>IMAGE_FILE_LEN_MAX) {
                        // File is "full"
                        diag_printf("TFTPD: image file too large(>%d)\n", IMAGE_FILE_LEN_MAX);
                        tftpd_send_error(reply,TFTP_ENOSPACE);     
                        ok = false;  // Give up
                        break; // out of the retries loop
                    }
                    memcpy((gImgBuf+gImgLen),response->th_data, (data_len-4));
                    gImgLen += (data_len-4);
                    
                    if (data_len < (SEGSIZE+4)) {
                        // End of file
                        closed = true;
                        ok = false;
                        
                        /*diag_printf("TFTPD: End of file\n");*/
                        diag_printf("total %d bytes received.\n", gImgLen);
                        
                        // Exception to the loop structure: we must ACK the last
                        // packet, the one that implied EOF:
    
                        reply->th_opcode = htons(ACK);
                        reply->th_block = htons(block++); // postincrement
                        __udp_sendto((char *)reply, 4, &tftpd_stream.cli_addr, &tftpd_stream.serv_addr);

                        if (1 == upgrade_firmware_main())
                            return 1; // success to upgarde firmware
                        
                        break; // out of the retries loop
                    }
                    // Happy!  Break out of the retries loop.
                    break;
                }
            }
            // Otherwise, we got something we do not understand!  So repeat
            // sending the current ACK, and use up a retry count.

        } // End of the retries loop.
        if (TFTP_RETRIES_MAX <= tries) {
            tftpd_send_error(reply,TFTP_EBADOP);
            ok = false;
        }
    }
    return 0; // fail to upgarde firmware
}

//
// Send a file to the client
//
static void 
tftpd_read_file(struct tftphdr *hdr)
{
    char data_out[SEGSIZE+sizeof(struct tftphdr)];
    char data_in[SEGSIZE+sizeof(struct tftphdr)];
    struct tftphdr *reply = (struct tftphdr *)data_out;
    struct tftphdr *response = (struct tftphdr *)data_in;
    int len, tries, ok, data_len;
    unsigned short block;
    struct timeval timeout;
    int total_timeouts = 0;
    char *flash_ptr;
    
    block = 0;
    ok = true;
    flash_ptr = FLASH_START_ADDR;
    while (ok) {
        // Read next chunk of file
        if ((block*SEGSIZE) >= FLASH_SIZE) {
            len = 0; /* the end of flash */
        }
        else {
            memcpy(reply->th_data, (flash_ptr+block*SEGSIZE), SEGSIZE);
            len = SEGSIZE;
        }
        reply->th_block = htons(++block); // preincrement
        reply->th_opcode = htons(DATA);
        for (tries = 0;  tries < TFTP_RETRIES_MAX;  tries++) {
            if (__udp_sendto((char *)reply, (4+len), &tftpd_stream.cli_addr, &tftpd_stream.serv_addr) < 0) {
                // Something went wrong with the network!
                diag_printf("TFTPD: error to send DATA\n");
                ok = false;
                break;
            }
        repeat_select:
            timeout.tv_sec = TFTP_TIMEOUT_PERIOD;
            timeout.tv_usec = 0;
            data_len = sizeof(data_in);
            data_len = __udp_recvfrom(data_in, data_len, &tftpd_stream.cli_addr,
                                       &tftpd_stream.serv_addr, &timeout);
            
            if ( data_len < 0) {
                diag_printf("TFTPD timed out 3\n");
                if (++total_timeouts > TFTP_TIMEOUT_MAX)
                {
                    tftpd_send_error(reply,TFTP_EBADOP);
                    ok = false;
                    break; // out of the retries loop
                }
                continue; // retry the send, using up one retry.
            }
            else {
                if ((block%128) == 0)
                    sysGpioSet(7, 0); /* turn off Test LED */
                else if ((block%64) == 0)
                    sysGpioSet(7, 1); /* turn on Test LED */
                /* 
                 * Print one dot character whenever receiving 1K bytes,
                 * and 64 dot characters (64K) per line.
                 */               
                if ((block%2) == 0)
                    diag_printf(".");
                if ((block%128) == 0)
                    diag_printf("\n");
                    
                if ((ntohs(response->th_opcode) == ACK) &&
                    (ntohs(response->th_block) < block)) {
    
                    // Then it is a repeat ACK for an old block; listen again,
                    // but do not repeat sending the current block, and do not
                    // use up a retry count.  (we do re-send the data if
                    // subsequently we time out)
                    goto repeat_select;
                }
                if ((ntohs(response->th_opcode) == ACK) &&
                    (ntohs(response->th_block) == block)) {
    
                    // Happy!  Break out of the retries loop.
                    break;
                }
                // Otherwise, we got something we do not understand!  So repeat
                // sending the current block, and use up a retry count.
            }
        } // End of the retries loop.
        if (TFTP_RETRIES_MAX <= tries) {
            tftpd_send_error(reply,TFTP_EBADOP);
            ok = false;
        }
        if (len < SEGSIZE) {
            diag_printf("total %d bytes sent.\n", ((block-1)*SEGSIZE)+len );
            break; // That's end of file then.
        }
    }
}

//
// Actual TFTP server
//
int tftpd_server(void)
{
    int data_len, count;
    struct timeval timeout;
    char data[SEGSIZE+sizeof(struct tftphdr)];
    struct tftphdr *hdr = (struct tftphdr *)data;
    
    /* init */
    tftpd_stream.open = false;
    tftpd_stream.packets_received = 0;    
    tftpd_stream.last_good_block = 0;
    tftpd_stream.avail = 0;
    tftpd_stream.actual_len = -1;
    tftpd_stream.cli_addr.sin_port = 0;    
    
    /* Bind our local address so that the client can send to us. */
    memset((char *)&tftpd_stream.serv_addr, 0, sizeof(tftpd_stream.serv_addr));
    tftpd_stream.serv_addr.sin_family = AF_INET;
    tftpd_stream.serv_addr.sin_len = sizeof(tftpd_stream.serv_addr);
    tftpd_stream.serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    tftpd_stream.serv_addr.sin_port = htons(TFTP_PORT); /* tftp server port = 69 */
    
    diag_printf("TFTP server is running.\n");
    count = 0;
    while(1) {
        tftpd_stream.cli_addr.sin_port = 0;
        
        /* Bind our local address so that the client can send to us. */
        memset((char *)&tftpd_stream.serv_addr, 0, sizeof(tftpd_stream.serv_addr));
        tftpd_stream.serv_addr.sin_family = AF_INET;
        tftpd_stream.serv_addr.sin_len = sizeof(tftpd_stream.serv_addr);
        tftpd_stream.serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        tftpd_stream.serv_addr.sin_port = htons(TFTP_PORT); /* tftp server port = 69 */
    
        //timeout.tv_sec = (tftpd_stream.last_good_block == 0) ? 10*TFTP_TIMEOUT_PERIOD : TFTP_TIMEOUT_PERIOD;
        //timeout.tv_usec = 0;
        timeout.tv_sec = 0;
        timeout.tv_usec = 250000; //250 ms
        data_len = sizeof(data);
        data_len = __udp_recvfrom(data, data_len, &tftpd_stream.cli_addr,
                                       &tftpd_stream.serv_addr, &timeout);
        if ( data_len < 0) {
            //diag_printf("TFTPD timed out 1\n");
            /* Blink TEST LED in the following period when tftp server is running.
             * LED on:0.25 sec, off:0.75 sec.
             */
            count++;
            if (count%4 >= 1)
                sysGpioSet(7, 0); /* turn off Test LED */
            else
                sysGpioSet(7, 1); /* turn on Test LED */
            
            check_reset_button(250); //250 ms
            
            continue;
        }
        else {
            /*diag_printf("TFTPD: received %x from %s:%d\n", 
                        ntohs(hdr->th_opcode), inet_ntoa(&tftpd_stream.cli_addr.sin_addr),
                        tftpd_stream.cli_addr.sin_port);
             */

            switch (ntohs(hdr->th_opcode)) {
            case WRQ:
                if(1 == tftpd_write_file(hdr))
                    return 1; // success to upgarde firmware
                /*diag_printf("tftpd_write_file()\n");*/
                break;
            case RRQ:
                if ((strncmp(hdr->th_stuff, "bye", 3)) == 0) {
                    tftpd_send_error(hdr,TFTP_ENOTFOUND);
                    return 0;
                }
                else if ((strncmp(hdr->th_stuff, "flash", 9)) == 0) {
                    tftpd_read_file(hdr);
                    /*diag_printf("tftpd_read_file()\n");*/
                }
                else
                    tftpd_send_error(hdr,TFTP_ENOTFOUND);
                break;
            case ACK:
            case DATA:
            case ERROR:
                // Ignore
                break;
            default:
                diag_printf("TFTPD: bogus request %x from %s:%d\n",
                            ntohs(hdr->th_opcode),
                            inet_ntoa(&tftpd_stream.cli_addr.sin_addr),
                            tftpd_stream.cli_addr.sin_port);
                tftpd_send_error(hdr,TFTP_EBADOP);
            }
        }
    }
    return 0;
}

// EOF tftp_server.c

/*************************************************************************************/
/*             Firmware upgrade routines                                             */
/*************************************************************************************/

unsigned long calculate_checksum (int action, char *s, int size)
{
    static unsigned long c0, c1;
    unsigned long checksum, b;
    int i;

    switch (action)
    {
    case 0:
        c0 = c1 = 0;
        break;
    case 1:
        for (i = 0; i < size; i++)
        {
            c0 += s[i] & 0xff;
            c1 += c0;
        }
        break;
    case 2:
        b = (c0 & 65535) + ((c0 >> 16) & 65535);
        c0 = ((b >> 16) + b) & 65535;

        b = (c1 & 65535) + ((c1 >> 16) & 65535);
        c1 = ((b >> 16) + b) & 65535;

        checksum = ((c1 << 16) | c0);

        return checksum;
    }
    return 0;
}

/*
 * Function called to upgrade firmware.
 */
int upgrade_firmware_main(void)
{
    int i, stat, len, board_ok;
    char *ptr;
    unsigned long fileChksum, calChksum, rootfs_chksum, kernel_chksum;
    unsigned long hdrLen, kernelLen, rootfsLen, imgLen;
    void *err_addr;
    char board_id[256];
    
    if (gImgBuf == NULL) {
        diag_printf("gImgBuf null\n");
        return 0;
    }
        
    /* 1. check magic number */
    ptr = gImgBuf; /* point to magic number field */
    if (strncmp(ptr, "*#$^", 4) != 0) {
        diag_printf("Incorrect magic number!\n");
        return 0;
    }
    
    /* 2. verify header checksum */
    ptr = gImgBuf + 4; /* point to header length field */
    hdrLen = (ptr[3] & 0x000000ff) |
             ((ptr[2] & 0x000000ff) << 8) |
             ((ptr[1] & 0x000000ff) << 16) |
             ((ptr[0] & 0x000000ff) << 24);
    /*diag_printf("hdrLen=%d\n", hdrLen);*/
    ptr = gImgBuf + 36; /* point to header checksum field */
    fileChksum = (ptr[3] & 0x000000ff) |
                 ((ptr[2] & 0x000000ff) << 8) |
                 ((ptr[1] & 0x000000ff) << 16) |
                 ((ptr[0] & 0x000000ff) << 24);
    /*diag_printf("header chksum=%X\n", fileChksum);*/
    /* fill in the header checksum field with zero before caculate checksum */
    ptr[0] = 0x00;
    ptr[1] = 0x00;
    ptr[2] = 0x00;
    ptr[3] = 0x00;
    /* caculate checksum */
    calculate_checksum (0, NULL, 0);
    calculate_checksum (1, gImgBuf, hdrLen);
    calChksum = calculate_checksum (2, NULL, 0);
    if (calChksum != fileChksum) {
        diag_printf("Header checksum error!\n");
        return 0;
    }
    /* 3. check board id */
    ptr = gImgBuf + 40; /* point to board id field */
    /*
    diag_printf("Board ID(file)=");
    for (i=0;i<(int)(hdrLen-40);i++) {
        diag_printf("%c", ptr[i]);
    }
    diag_printf("\nlen=%d\n",(hdrLen-40));
    */
    board_ok = 0;
    memcpy(board_id, (char *)BOARD_ID_FLASH_ADDR, 256);
    board_id[255] = '\0';
    /*diag_printf("Board ID(flash)=%s\n", board_id);*/
    len = strlen(board_id); /* Board ID in flash contains NULL character. */
    for (i=0; i<(int)(hdrLen-40-len+1); i++) {
        if (strncmp(board_id, ptr, len) == 0) {
            board_ok = 1;
            break;
        }
        else
            ptr++;
    }
    if (board_ok == 0) {
        diag_printf("Board ID not matched!\n");
        return 0;
    }
    
    /* 4. verify image length */
    ptr = gImgBuf + 24; /* point to kernel length field */
    kernelLen = (ptr[3] & 0x000000ff) |
                ((ptr[2] & 0x000000ff) << 8) |
                ((ptr[1] & 0x000000ff) << 16) |
                ((ptr[0] & 0x000000ff) << 24);
    /*diag_printf("kernelLen=%X(%d)\n", kernelLen, kernelLen);*/
    ptr = gImgBuf + 28; /* point to root fs length field */
    rootfsLen = (ptr[3] & 0x000000ff) |
                ((ptr[2] & 0x000000ff) << 8) |
                ((ptr[1] & 0x000000ff) << 16) |
                ((ptr[0] & 0x000000ff) << 24);
    /*diag_printf("rootfsLen=%X(%d)\n", rootfsLen, rootfsLen);*/

    if (kernelLen > (KERNEL_FLASH_SIZE-8)) {
        diag_printf("kernel image > %d bytes!\n", (KERNEL_FLASH_SIZE-8));
        return 0;
    }
    else if (rootfsLen > (ROOTFS_FLASH_SIZE-8)) {
        diag_printf("rootfs image > %d bytes!\n", (ROOTFS_FLASH_SIZE-8));
        return 0;
    }

    /* 5. verify kernel checksum */
    if (kernelLen > 0) {
        ptr = gImgBuf + 16; /* point to kernel checksum field */
        fileChksum = (ptr[3] & 0x000000ff) |
                     ((ptr[2] & 0x000000ff) << 8) |
                     ((ptr[1] & 0x000000ff) << 16) |
                     ((ptr[0] & 0x000000ff) << 24);
        /*diag_printf("kernel chksum=%X\n", fileChksum);*/
        /* caculate checksum */
        ptr = gImgBuf + hdrLen + rootfsLen; /* point to kernel image */
        calculate_checksum (0, NULL, 0);
        calculate_checksum (1, ptr, kernelLen);
        kernel_chksum = calculate_checksum (2, NULL, 0);
        if (kernel_chksum != fileChksum) {
            diag_printf("kernel checksum error!\n");
            return 0;
        }
    }

    /* 6. verify rootfs checksum */
    if (rootfsLen > 0) {
        ptr = gImgBuf + 20; /* point to rootfs checksum field */
        fileChksum = (ptr[3] & 0x000000ff) |
                     ((ptr[2] & 0x000000ff) << 8) |
                     ((ptr[1] & 0x000000ff) << 16) |
                     ((ptr[0] & 0x000000ff) << 24);
        /*diag_printf("rootfs chksum=%X\n", fileChksum);*/
        /* caculate checksum */
        ptr = gImgBuf + hdrLen; /* point to rootfs image */
        calculate_checksum (0, NULL, 0);
        calculate_checksum (1, ptr, rootfsLen);
        rootfs_chksum = calculate_checksum (2, NULL, 0);
        if (rootfs_chksum != fileChksum) {
            diag_printf("rootfs checksum error!\n");
            return 0;
        }
    }
    
    /* 7. verify image checksum */
    ptr = gImgBuf + 32; /* point to image checksum field */
    fileChksum = (ptr[3] & 0x000000ff) |
                 ((ptr[2] & 0x000000ff) << 8) |
                 ((ptr[1] & 0x000000ff) << 16) |
                 ((ptr[0] & 0x000000ff) << 24);
    /*diag_printf("image chksum=%X\n", fileChksum);*/
    
    imgLen = kernelLen + rootfsLen;
    ptr = gImgBuf + hdrLen; /* point to the start location of image body (rootfs image) */
    calculate_checksum (0, NULL, 0);
    calculate_checksum (1, ptr, imgLen);
    calChksum = calculate_checksum (2, NULL, 0);
    if (calChksum != fileChksum) {
        diag_printf("Image checksum error!\n");
        return 0;
    }
    
    /* 8. write to flash and reboot */
    /* upgrade rootfs image */
    write_to_flash_ongoing = 1;
    if (rootfsLen > 0) {
        // Safety check - make sure the address range is not within the code we're running
        if (flash_code_overlaps((void *)ROOTFS_FLASH_ADDR, (void *)(ROOTFS_FLASH_ADDR+ROOTFS_FLASH_SIZE-1))) {
            diag_printf("Can't program this region - contains code in use!\n");
            write_to_flash_ongoing = 0;
            return 0;
        }
        diag_printf("Start to upgrade rootfs image.\n");
        // Erase area to be programmed
        if ((stat = flash_erase((void *)ROOTFS_FLASH_ADDR, ROOTFS_FLASH_SIZE, (void **)&err_addr)) != 0) {
            diag_printf("Can't erase region at %p: %s\n", err_addr, flash_errmsg(stat));
            write_to_flash_ongoing = 0;
            return 0;
        }
        // Now program it
        ptr = gImgBuf + hdrLen; /* point to rootfs image */
        memset(gTempImgBuf, '\0', ROOTFS_FLASH_SIZE);
        memcpy(gTempImgBuf, ptr, rootfsLen); /* have a copy to avoid word alignment iisue */
        memcpy((gTempImgBuf+ROOTFS_FLASH_SIZE-8), (gImgBuf+28), 4); /* rootfs length */
        memcpy((gTempImgBuf+ROOTFS_FLASH_SIZE-4), (gImgBuf+20), 4); /* rootfs checksum */
        if ((stat = FLASH_PROGRAM((void *)ROOTFS_FLASH_ADDR, (void *)gTempImgBuf, ROOTFS_FLASH_SIZE, (void **)&err_addr)) != 0) {
            diag_printf("Can't program region at %p: %s\n", err_addr, flash_errmsg(stat));
            write_to_flash_ongoing = 0;
            return 0;
        }
    }
    
    /* upgrade kernel image */
    if (kernelLen > 0) {
        // Safety check - make sure the address range is not within the code we're running
        if (flash_code_overlaps((void *)KERNEL_FLASH_ADDR, (void *)(KERNEL_FLASH_ADDR+KERNEL_FLASH_SIZE-1))) {
            diag_printf("Can't program this region - contains code in use!\n");
            write_to_flash_ongoing = 0;
            return 0;
        }
        diag_printf("Start to upgrade kernel image.\n");
        // Erase area to be programmed
        if ((stat = flash_erase((void *)KERNEL_FLASH_ADDR, KERNEL_FLASH_SIZE, (void **)&err_addr)) != 0) {
            diag_printf("Can't erase region at %p: %s\n", err_addr, flash_errmsg(stat));
            write_to_flash_ongoing = 0;
            return 0;
        }
        // Now program it
        ptr = gImgBuf + hdrLen + rootfsLen; /* point to kernel image */
        memset(gTempImgBuf, '\0', KERNEL_FLASH_SIZE);
        memcpy(gTempImgBuf, ptr, kernelLen); /* have a copy to avoid word alignment iisue */
        memcpy((gTempImgBuf+KERNEL_FLASH_SIZE-8), (gImgBuf+24), 4); /* kernel length */
        memcpy((gTempImgBuf+KERNEL_FLASH_SIZE-4), (gImgBuf+16), 4); /* kernel checksum */
        if ((stat = FLASH_PROGRAM((void *)KERNEL_FLASH_ADDR, (void *)gTempImgBuf, KERNEL_FLASH_SIZE, (void **)&err_addr)) != 0) {
            diag_printf("Can't program region at %p: %s\n", err_addr, flash_errmsg(stat));
            write_to_flash_ongoing = 0;
            return 0;
        }
    }
    
    diag_printf("Upgrade completely.\n");
    write_to_flash_ongoing = 0;
    return 1;
}

int kernel_rootfs_image_integrity_check(void)
{
    unsigned long img_chksum, cal_chksum;
    unsigned long img_len;
    char *ptr;
    
    /* kernel checksum */
    ptr = KERNEL_FLASH_ADDR + KERNEL_FLASH_SIZE - 8; /* point to kernel length field */
    img_len = (ptr[3] & 0x000000ff) |
              ((ptr[2] & 0x000000ff) << 8) |
              ((ptr[1] & 0x000000ff) << 16) |
              ((ptr[0] & 0x000000ff) << 24);
    /*diag_printf("kernelLen=%X(%d)\n", img_len, img_len);*/
    
    if (img_len > 0) {
        ptr = KERNEL_FLASH_ADDR + KERNEL_FLASH_SIZE - 4; /* point to kernel checksum field */
        img_chksum = (ptr[3] & 0x000000ff) |
                     ((ptr[2] & 0x000000ff) << 8) |
                     ((ptr[1] & 0x000000ff) << 16) |
                     ((ptr[0] & 0x000000ff) << 24);
        /*diag_printf("kernel chksum=%X\n", img_chksum);*/
        /* caculate checksum */
        ptr = KERNEL_FLASH_ADDR;
        calculate_checksum (0, NULL, 0);
        calculate_checksum (1, ptr, img_len);
        cal_chksum = calculate_checksum (2, NULL, 0);
        if (cal_chksum != img_chksum) {
            diag_printf("kernel checksum error!\n");
            return 0;
        }
        else
            diag_printf("kernel checksum ok.\n");
    }

    /* rootfs checksum */
    ptr = ROOTFS_FLASH_ADDR + ROOTFS_FLASH_SIZE - 8; /* point to kernel length field */
    img_len = (ptr[3] & 0x000000ff) |
              ((ptr[2] & 0x000000ff) << 8) |
              ((ptr[1] & 0x000000ff) << 16) |
              ((ptr[0] & 0x000000ff) << 24);
    /*diag_printf("rootfsLen=%X(%d)\n", img_len, img_len);*/
    
    if (img_len > 0) {
        ptr = ROOTFS_FLASH_ADDR + ROOTFS_FLASH_SIZE - 4; /* point to kernel checksum field */
        img_chksum = (ptr[3] & 0x000000ff) |
                     ((ptr[2] & 0x000000ff) << 8) |
                     ((ptr[1] & 0x000000ff) << 16) |
                     ((ptr[0] & 0x000000ff) << 24);
        /*diag_printf("rootfs chksum=%X\n", img_chksum);*/
        /* caculate checksum */
        ptr = ROOTFS_FLASH_ADDR;
        calculate_checksum (0, NULL, 0);
        calculate_checksum (1, ptr, img_len);
        cal_chksum = calculate_checksum (2, NULL, 0);
        if (cal_chksum != img_chksum) {
            diag_printf("rootfs checksum error!\n");
            return 0;
        }
        else
            diag_printf("rootfs checksum ok.\n");
    }
    return 1;
}

/*
void burn_kernel_rootfs_len_chksum(int type, unsigned long len, unsigned long chksum)
{
    char buf[8], *ptr;
    int stat;
    void *err_addr;
    
    if (type == 1) {
        ptr = (char*)&len; //kernel length
        memcpy(buf, ptr, 4);
        ptr = (char*)&chksum; //kernel checksum
        memcpy(buf+4, ptr, 4);
        if ((stat = FLASH_PROGRAM((void *)(KERNEL_FLASH_ADDR+KERNEL_FLASH_SIZE-8), (void *)buf, 8, (void **)&err_addr)) != 0) {
                diag_printf("Can't program region at %p: %s\n", err_addr, flash_errmsg(stat));
                return 0;
        }
    }
    else if (type == 2)
    {
        ptr = (char*)&len; //rootfs length
        memcpy(buf, ptr, 4);
        ptr = (char*)&chksum; //rootfs checksum
        memcpy(buf+4, ptr, 4);
        if ((stat = FLASH_PROGRAM((void *)(ROOTFS_FLASH_ADDR+ROOTFS_FLASH_SIZE-8), (void *)buf, 8, (void **)&err_addr)) != 0) {
                diag_printf("Can't program region at %p: %s\n", err_addr, flash_errmsg(stat));
                return 0;
        }
    }
}
*/
